use std::fmt;
use std::convert::TryFrom;
use std::net::{Ipv6Addr, Ipv4Addr};

use serde::{Deserialize, Serialize};

use trust_dns_client::serialize::binary::BinEncoder;
use trust_dns_proto::error::ProtoError;

use super::trust_dns_types;
use super::name::SerdeName;


#[derive(Deserialize, Serialize, Clone)]
#[serde(tag = "Type")]
#[serde(rename_all = "UPPERCASE")]
pub enum RData {
    #[serde(rename_all = "PascalCase")]
    A {
        address: Ipv4Addr
    },
    #[serde(rename_all = "PascalCase")]
    AAAA {
        address: Ipv6Addr
    },
    #[serde(rename_all = "PascalCase")]
    CAA {
        issuer_critical: bool,
        value: String,
        property_tag: String,
    },
    #[serde(rename_all = "PascalCase")]
    CNAME {
        target: SerdeName
    },
    // HINFO(HINFO),
    // HTTPS(SVCB),
    #[serde(rename_all = "PascalCase")]
    MX {
        preference: u16,
        mail_exchanger: SerdeName
    },
    // NAPTR(NAPTR),
    #[serde(rename_all = "PascalCase")]
    NULL {
        data: String
    },
    #[serde(rename_all = "PascalCase")]
    NS {
        target: SerdeName
    },
    // OPENPGPKEY(OPENPGPKEY),
    // OPT(OPT),
    #[serde(rename_all = "PascalCase")]
    PTR {
        target: SerdeName
    },
    #[serde(rename_all = "PascalCase")]
    SOA {
        master_server_name: SerdeName,
        maintainer_name: SerdeName,
        refresh: i32,
        retry: i32,
        expire: i32,
        minimum: u32,
        serial: u32
    },
    #[serde(rename_all = "PascalCase")]
    SRV {
        server: SerdeName,
        port: u16,
        priority: u16,
        weight: u16,
    },
    #[serde(rename_all = "PascalCase")]
    SSHFP {
        algorithm: u8,
        digest_type: u8,
        fingerprint: String,
    },
    // SVCB(SVCB),
    // TLSA(TLSA),
    #[serde(rename_all = "PascalCase")]
    TXT {
        text: String
    },

    // TODO: Eventually allow deserialization of DNSSEC records
    #[serde(skip)]
    DNSSEC(trust_dns_types::DNSSECRData),
    #[serde(rename_all = "PascalCase")]
    Unknown {
        code: u16,
        data: String,
    },
    // ZERO,

    // TODO: DS
    // TODO: TLSA
}

impl From<trust_dns_types::RData> for RData {
    fn from(rdata: trust_dns_types::RData) -> RData {
        match rdata {
            trust_dns_types::RData::A(address) => RData::A { address },
            trust_dns_types::RData::AAAA(address) => RData::AAAA { address },
            // Still a draft, no iana number yet, I don't to put something that is not currently supported so that's why NULL and not unknown.
            // TODO: probably need better error here, I don't know what to do about that as this would require to change the From for something else.
            // (empty data because I'm lazy)
            trust_dns_types::RData::ANAME(_) =>  RData::NULL {
                data: String::new()
            },
            trust_dns_types::RData::CNAME(target) => RData::CNAME {
                target: SerdeName(target)
            },
            trust_dns_types::RData::CAA(caa) => RData::CAA {
                issuer_critical: caa.issuer_critical(),
                value: format!("{}", CAAValue(caa.value())),
                property_tag: caa.tag().as_str().to_string(),
            },
            trust_dns_types::RData::MX(mx) => RData::MX {
                preference: mx.preference(),
                mail_exchanger: SerdeName(mx.exchange().clone())
            },
            trust_dns_types::RData::NULL(null) =>  RData::NULL {
                data: base64::encode(null.anything().map(|data| data.to_vec()).unwrap_or_default())
            },
            trust_dns_types::RData::NS(target) => RData::NS {
                target: SerdeName(target)
            },
            trust_dns_types::RData::PTR(target) => RData::PTR {
                target: SerdeName(target)
            },
            trust_dns_types::RData::SOA(soa) => RData::SOA {
                master_server_name: SerdeName(soa.mname().clone()),
                maintainer_name: SerdeName(soa.rname().clone()),
                refresh: soa.refresh(),
                retry: soa.retry(),
                expire: soa.expire(),
                minimum: soa.minimum(),
                serial: soa.serial()
            },
            trust_dns_types::RData::SRV(srv) => RData::SRV {
                server: SerdeName(srv.target().clone()),
                port: srv.port(),
                priority: srv.priority(),
                weight: srv.weight(),
            },
            trust_dns_types::RData::SSHFP(sshfp) => RData::SSHFP {
                algorithm: sshfp.algorithm().into(),
                digest_type: sshfp.fingerprint_type().into(),
                fingerprint: trust_dns_types::sshfp::HEX.encode(sshfp.fingerprint()),
            },
            //TODO: This might alter data if not utf8 compatible, probably need to be replaced
            //TODO: check whether concatenating txt data is harmful or not
            trust_dns_types::RData::TXT(txt) => RData::TXT { text: format!("{}", txt) },
            trust_dns_types::RData::DNSSEC(data) => RData::DNSSEC(data),
            rdata => {
                let code = rdata.to_record_type().into();
                let mut data = Vec::new();
                let mut encoder = BinEncoder::new(&mut data);
                // TODO: need better error handling (use TryFrom ?)
                rdata.emit(&mut encoder).expect("could not encode data");

                RData::Unknown {
                    code,
                    data: base64::encode(data),
                }
            }
        }
    }
}

impl TryFrom<RData> for trust_dns_types::RData {
    type Error = ProtoError;

    fn try_from(rdata: RData) -> Result<Self, Self::Error> {
        Ok(match rdata {
            RData::A { address } => trust_dns_types::RData::A(address),
            RData::AAAA { address } => trust_dns_types::RData::AAAA(address),
            // TODO: Round trip test all types below (currently not tested...)
            RData::CAA { issuer_critical, value, property_tag } => {
                let property = trust_dns_types::caa::Property::from(property_tag);
                let caa_value = {
                    // TODO: duplicate of trust_dns_client::serialize::txt::rdata_parser::caa::parse
                    // because caa::read_value is private
                    match property {
                        trust_dns_types::caa::Property::Issue | trust_dns_types::caa::Property::IssueWild => {
                            let value = trust_dns_types::caa::read_issuer(value.as_bytes())?;
                            trust_dns_types::caa::Value::Issuer(value.0, value.1)
                        }
                        trust_dns_types::caa::Property::Iodef => {
                            let url = trust_dns_types::caa::read_iodef(value.as_bytes())?;
                            trust_dns_types::caa::Value::Url(url)
                        }
                        trust_dns_types::caa::Property::Unknown(_) => trust_dns_types::caa::Value::Unknown(value.as_bytes().to_vec()),
                    }
                };
                trust_dns_types::RData::CAA(trust_dns_types::caa::CAA {
                    issuer_critical,
                    tag: property,
                    value: caa_value,
                })
            },
            RData::CNAME { target } => trust_dns_types::RData::CNAME(target.into_inner()),
            RData::MX { preference, mail_exchanger } => trust_dns_types::RData::MX(
                trust_dns_types::mx::MX::new(preference, mail_exchanger.into_inner())
            ),
            RData::NULL { data } => trust_dns_types::RData::NULL(
                trust_dns_types::null::NULL::with(
                    base64::decode(data).map_err(|e| ProtoError::from(format!("{}", e)))?
                )
            ),
            RData::NS { target } => trust_dns_types::RData::NS(target.into_inner()),
            RData::PTR { target } => trust_dns_types::RData::PTR(target.into_inner()),
            RData::SOA {
                master_server_name,
                maintainer_name,
                refresh,
                retry,
                expire,
                minimum,
                serial
            } => trust_dns_types::RData::SOA(
                trust_dns_types::soa::SOA::new(
                    master_server_name.into_inner(),
                    maintainer_name.into_inner(),
                    serial,
                    refresh,
                    retry,
                    expire,
                    minimum,
                )
            ),
            RData::SRV { server, port, priority, weight } => trust_dns_types::RData::SRV(
                trust_dns_types::srv::SRV::new(priority, weight, port, server.into_inner())
            ),
            RData::SSHFP { algorithm, digest_type, fingerprint } => trust_dns_types::RData::SSHFP(
                trust_dns_types::sshfp::SSHFP::new(
                    // NOTE: This allows unassigned algorithms
                    trust_dns_types::sshfp::Algorithm::from(algorithm),
                    trust_dns_types::sshfp::FingerprintType::from(digest_type),
                    trust_dns_types::sshfp::HEX.decode(fingerprint.as_bytes()).map_err(|e| ProtoError::from(format!("{}", e)))?
                )
            ),
            RData::TXT { text } => trust_dns_types::RData::TXT(trust_dns_types::txt::TXT::new(vec![text])),
            // TODO: Error out for DNSSEC? Prefer downstream checks?
            RData::DNSSEC(_) => todo!(),
            // TODO: Disallow unknown? (could be used to bypass unsopported types?) Prefer downstream checks?
            RData::Unknown { code, data } => todo!(),
        })
    }
}

struct CAAValue<'a>(&'a trust_dns_types::caa::Value);

// trust_dns Display implementation panics if no parameters
// Implementation based on caa::emit_value
// Also the quotes are strips to render in JSON
impl<'a> fmt::Display for CAAValue<'a> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
        match self.0 {
            trust_dns_types::caa::Value::Issuer(name, parameters) => {

                if let Some(name) = name {
                    write!(f, "{}", name)?;
                }

                if name.is_none() && parameters.is_empty() {
                    write!(f, ";")?;
                }

                for value in parameters {
                    write!(f, "; {}", value)?;
                }
            }
            trust_dns_types::caa::Value::Url(url) => write!(f, "{}", url)?,
            trust_dns_types::caa::Value::Unknown(v) => write!(f, "{:?}", v)?,
        }
        Ok(())
    }
}
